# DESCRIPTION
# This is a Perl script that attempts to convert ABC music notation
# containing chords to the MMA format. Melodies are imported via MMA's
# MIDIinc command, assuming abc2midi was used to generate the midi file.

# FILENAME
# abc2mma.pl

# AUTHOR
# Gek S. Low

# INSTRUCTIONS

$ver = "0.51";

$help_text = "
abc2mma.pl version $ver.
Copyright (c) Gek S. Low, 2006.

Usage:
    abc2mma.pl <abcfile.abc | abcfile.txt> [options]

    output filename is in the form: <abcfile><N><suffix>.mma
	 
	 - default=<abcfile><N>_mma.mma
    - the _mma suffix is added to avoid name collision with
      abc2midi ouput
    - N is the song reference number (X:) in the ABC file

Options:

-o or -output=<mmafile>
    - output to <mmafile>.mma (overrides -songref and -suffix)
-songref=<yes|no>
    - sets whether to use the song reference number in outfile filename
      also affects the MIDI filename in the MIDIinc line
      (default=yes)
-suffix=<suffix>
    - appends suffix to filename (default=_mma)
-writefile=<yes|no>
    - generates file (overrides -outfile) (default=yes)
-includemidi=<yes|no>
    - inserts MIDIinc command to include <abcfile>.mid (default=yes)
-groove=<mma_groove>
    - sets Groove (default=Folk)
-swingmode=<yes|no>
    - sets SwingMode (default=yes)
-tempo=<t>
    - sets tempo (default=80)\n";

# NOTES / DISCLAIMERS
#
# - This is NOT a parser! I'm just using mostly string substitution to get
# the results I want. I'm too lazy to write a full abc parser.
#
# - I do not claim to handle all possible abc notation scenarios. There are
# features of abc notation which I don't use, so I'm sure there are many
# cases which are broken.
#
# - Finally, use at your own risk! I'm not responsible for any damages, blah
# blah... that may result from the use of this software.

# TO DO LIST
# - Parts
# - Voices and multiple staffs
# - Lyrics

# CHANGE LOG
# Version 0.5: 2 Apr 2006
# Version 0.51: 14 Apr 2006 - Fixed logical bug in compound time

# CONTACT
# gek<remove this>siong AT <remove this - no spam!>yahoo DOT com
# This file and contents is copyright (c) Gek S. Low 2006

sub ord_str() {
  my @ret = unpack("c*", shift);
  my $ret = join(' ',@ret);
  return $ret;
}

sub chr_str() {
  my @ret = split(" ", shift);
  my $ret = pack("c*", @ret);
  return $ret;
}

sub calc_strums() {
  my $str = shift;
  my $count = eval($str) / ($note_length / $meter_bottom);
  my $ret="";
  for ($i=0; $i<$count; $i++) {
    $ret = $ret." /";
  }
  return $ret;
}


# Set up defaults
$mmafile = "";
$writefile = "yes";
$includemidi = "yes";
$file_songref = "yes";
$file_suffix = "_mma";

$abcfile = $ARGV[0];

if ($ARGV[0] eq "") {
  print $help_text;
  exit(0);
}

foreach $arg (@ARGV) {
  if ($arg =~ /\-o(utput)?\=(.*)$/) {
	 $mmafile = $2;
  }
  if ($arg =~ /\-suffix\=(.*)$/) {
	 $file_suffix = $1;
  }
  if ($arg =~ /\-songref\=(.*)$/) {
	 $file_songref = $1;
  }
  if ($arg =~ /\-writefile\=(.*)$/) {
	 $writefile = $1;
  }
  if ($arg =~ /\-swingmode\=(.*)$/) {
	 $swingmode = $1;
  }
  if ($arg =~ /\-groove\=(.*)$/) {
	 $groove = $1;
  }
  if ($arg =~ /\-tempo\=(.*)$/) {
	 $tempo = $1;
  }
  if ($arg =~ /\-includemidi\=(.*)$/) {
	 $includemidi = $1;
  }
}

open(INPUT, $abcfile) or die "Unable to open $abcfile.\n";

@input = <INPUT>;
$input = join('', @input);

close(INPUT);

$input =~ s/^(X:.*\n)/####$1/mg;
@input = split("####", $input);


foreach $data (@input) {
  # reset per-song defaults
  $meter = "1/4";
  $keysig = "C";
  $meter_top = 1; $meter_bottom = 4;
  $note_length = 4;

  $groove = "";
  $swingmode = "";
  $tempo = "";
  $songref = "";

  if ($data =~ /X:\s*(.*)$/m) {
	 if ($file_songref eq "yes") {
		$songref = $1;
	 }
	 
	 # Get title and clear it
	 if ($data =~ s/^T:\s*?(\S.*?)$//m) {
		$title = $1;
    } else { 
		$title = "Unknown";
	 }
	 
	 # Get meter and clear it
	 if ($data =~ s/^M:\s*?(\S.*?)$//m) { $meter = $1; } else { $meter = "4/4"; }

	 $meter =~ /(\d+)\/(\d+)/;
	 $meter_top = $1;
	 $meter_bottom = $2;
	 
	 # Get the base note length
	 if ($data =~ s/^L:\s*?1\/(\S.*?)$//m) {
		$note_length = $1;
	 } else {
		if ($meter_top / $meter_bottom < 0.75) {
		  $note_length = 16;
		} else {
		  $note_length = 8;
		}
	 }
	 
	 # Get tempo
	 if ($tempo eq "") {
		$tempo = 80;
		if ($data =~ s/^Q:.*?=\s*?(\d*?)$//m) { $tempo = $1; }
	 }
	 
	 # Get key signature
	 if ($data =~ s/^K:\s*?(\S.*?)$//m) {
		$keysig = $1;
		$keysig =~ s/m/ minor/;
		$keysig =~ s/\s?(treble[+-]?\d?|bass\d?|alto\d?|none|perc)//;
	 }
	 
	 # Get MMA-specific commands
	 if ($groove eq "") {
		$groove = "Folk";
		if ($data =~ s/%%X\-MMA Groove (.*?)$//m) { $groove = $1; }
	 }

	 if ($swingmode eq "") {
		$swingmode = "Off";
		if ($data =~ s/%%X\-MMA SwingMode (.*?)$//m) { $swingmode = $1; }
	 }

	 $data =~ s/^%%/####/mg;	# separate the commands from the comments
	 
	 $data =~ s/%.*?$//mg;		# clear out all comments
	 $data =~ s/\\//mg;
	 #$data =~ s/\\.*?\n//mg;		# join continuation lines \
	 $data =~ s/\"(.*?)\"/ \"$1\" /mg;	# clarify which are the quoted strings
	 $data =~ s/\"[\^_<>\@](.*?)\"/\"X$1\"/mg;	# ensure annotations are never mistaken for chords
	 $data =~ s/[-<>\.~]//mg;		# remove - < > . ~
	 $data =~ s/!.*?!//mg;		# remove !symbols!
	 $data =~ s/{.*?}//mg;		# remove graces {}
	 $data =~ s/\(([^\(]*?)\)/$1/mg;		# "de"-slur
	 $data =~ s/\[([\^_=]*?[A-Ga-gzx][\,']*).*?\]/$1/mg;	# "de"-chord

	 $data =~ s/\n+/\n/mg;	# remove blank lines prior to processing
	 
	 # deal with tuplets
	 $data =~ s/\(2\s*?([\^_=]*?[A-Ga-gzx][\,']*(\d?\/*)){2}/x$2x$2x$2/g;
	 $data =~ s/\(3\s*?([\^_=]*?[A-Ga-gzx][\,']*(\d?\/*)){3}/x$2x$2/g; # (3ABC
	 $data =~ s/\(4\s*?([\^_=]*?[A-Ga-gzx][\,']*(\d?\/*)){4}/x$2x$2x$2/g;
	 $data =~ s/\(6\s*?([\^_=]*?[A-Ga-gzx][\,']*(\d?\/*)){6}/x$2x$2/g;
	 $data =~ s/\(8\s*?([\^_=]*?[A-Ga-gzx][\,']*(\d?\/*)){8}/x$2x$2x$2/g;
	 
	 # deal with 5, 7 and 9 - determine what is n
	 if ($meter_top == 6 || $meter_top == 9 || $meter_top == 12) {
		# time is compound - n=3
		$data =~ s/\(5\s*?([\^_=]*?[A-Ga-gzx][\,']*(\d?\/*)){5}/x$2x$2x$2/g;
		$data =~ s/\(7\s*?([\^_=]*?[A-Ga-gzx][\,']*(\d?\/*)){7}/x$2x$2x$2/g;
		$data =~ s/\(9\s*?([\^_=]*?[A-Ga-gzx][\,']*(\d?\/*)){9}/x$2x$2x$2/g;
	 } else {
		  # time is simple - n=2
		  $data =~ s/\(5\s*?([\^_=]*?[A-Ga-gzx][\,']*(\d?\/*)){5}/x$2x$2/g;
		  $data =~ s/\(7\s*?([\^_=]*?[A-Ga-gzx][\,']*(\d?\/*)){7}/x$2x$2/g;
		  $data =~ s/\(9\s*?([\^_=]*?[A-Ga-gzx][\,']*(\d?\/*)){9}/x$2x$2/g;
	 }
	 
	 #print $data;
	 
	 # Start processing line by line
	 @data = split("\n", $data);

	 $output = "";

	 foreach $line (@data) {
		if ($line =~ /^\w:/) {
		  # we'll deal with lyrics, voices and parts in future versions
		} elsif ($line =~ /^####/) {
		  if ($line =~ /MMA (.*)$/) {
			 $mma_command = $1;
			 $output = $output."#MMA ".$mma_command."\n";
		  }
		} else {
		  # This should be a measure line

		  # protect the chords from the subsequent ops (clashes with note names)
		  $line =~ s/\"([A-G]([b#]?(m|min|maj|M|\+|#|dim|sus|\d).*?)?)\"/"!".&ord_str($1)."!"/ge;

		  # delete annotations that are not chords
		  $line =~ s/\".*?\"//g;
		  
		  # count the beats (handles single line melodies only), up to 32nd notes
		  $line =~ s/[\^_=]*?[A-Ga-gzx][\,']*\/\/\// +0.125 /g;
		  $line =~ s/[\^_=]*?[A-Ga-gzx][\,']*\/\// +0.25 /g;
		  $line =~ s/[\^_=]*?[A-Ga-gzx][\,']*\// +0.5 /g;
		  $line =~ s/[\^_=]*?[A-Ga-gzx][\,']*(\d)/ +$1 /g;
		  $line =~ s/[\^_=]*?[A-Ga-gzx][\,']*/ +1 /g;
		  
		  # put the chords back
		  $line =~ s/!(.*?)!/"\"".&chr_str($1)."\""/ge;
		  
		  # calculate the number of strums
		  $line =~ s/\s([\+\-\d.\s]+)/&calc_strums($1)." "/ge;
		  
		  $line =~ s/\s+/ /g;		# remove extra spaces
		  $line =~ s/" \//"/g;	# minus one strum for each chord present
		  $line =~ s/"(.*?)"/$1/g;	# unquote the chords
		  
		  $output = $output." ".$line;
		}
	 }
	 
	 # remove extra barlines
	 $output =~ s/\|[\s\n]*\|/\|/g;
	 
	 # convert the repeats and end
	 $output =~ s/\|:/\| #MMA Repeat\n/g;
	 $output =~ s/:\|(\[\d)?/\| #MMA RepeatEnd\n/g;
	 $output =~ s/\[\d/#MMA RepeatEnding\n/g;
	 $output =~ s/\|\]/\n/g;
	 
	 #print $output;
	 
	 # split by numbers
	 $output =~ s/\|/\n@/mg;
	 @measures = split("@", $output);
	 
	 # check for a pick-up bar, can remove
	 if ($measures[0] =~ /^\s*\//) {
		shift(@measures);
	 }
	 
	 # assign measure numbers
	 $measure_count = 1;
	 $mma_command = "";
	 for ($i=0; $i<scalar(@measures); $i++) {
		$measures[$i] =~ s/^\s+//;
		$measures[$i] = ($measure_count)."\t".$measures[$i];
		
		# if there's an MMA command, put it before the measure
		$orig_measure = $measures[$i];
		while ($orig_measure =~ /#MMA (.*)\n/) {
		  $mma_command = $1;
		  $orig_measure =~ s/#MMA $mma_command\s*//;
		  $measures[$i] =~ s/#MMA $mma_command\s*//;
		  $measures[$i] = $mma_command."\n".$measures[$i];
		}
		$measure_count++;
	 }
	 
	 $output = join("", @measures);
	 
	 # form the output filename
	 $outfile = $abcfile;
	 $outfile =~ s/\.\w\w\w$//;
	 $abcmidifile = $outfile.$songref.".mid";
	 if ($mmafile eq "") {
		if ($songref eq "no") {
		  $outfile = $outfile.$file_suffix.".mma";
		} else {
		  $outfile = $outfile.$songref.$file_suffix.".mma";
		}
	 } else {
		$outfile = $mmafile.".mma";
	 }

	 if ($writefile eq "yes") {
		open (OUTPUT, ">$outfile") or die "Unable to create $mmafile.\n";
	 } else {
		*OUTPUT = *STDOUT;
	 }

	 print OUTPUT "// $title\n";
	 print OUTPUT "// Created by abc2mma.pl ver $ver\n\n";
	 print OUTPUT "Groove $groove\n";
	 print OUTPUT "SwingMode $swingmode\n\n";
	 print OUTPUT "KeySig $keysig\n";
	 print OUTPUT "TimeSig $meter_top $meter_bottom\n\n";
	 print OUTPUT "Tempo $tempo\n\n";

	 $abcmidifile = $abcfile;
	 $abcmidifile =~ s/\.abc$/$songref\.mid/;

	 if ($includemidi eq "yes") {
		print OUTPUT "MIDIinc file=$abcmidifile Solo-Song=1 Drum=10 Volume=120\n";
	 }
	 print OUTPUT $output;

	 if ($writefile eq "yes") {
		print "Created $outfile.\n";
	 }

  }
}
